//
// See the file LICENSE for redistribution information.
//
// Copyright (c) 2002-2004
//	Sleepycat Software.  All rights reserved.
//
// $Id: DbXmlUri.cpp,v 1.5 2004/12/03 21:25:07 gmf Exp $
//

#include <dbxml/XmlPortability.hpp>
#include "DbXmlUri.hpp"
#include "../UTF8.hpp"
#include "../Globals.hpp"
#include "../Manager.hpp"
#include "../OperationContext.hpp"
#include "../scoped_ptr.hpp"
#include <dbxml/XmlContainer.hpp>
#include <dbxml/XmlTransaction.hpp>
#include <dbxml/XmlException.hpp>

#include <xercesc/util/XMLUri.hpp>

#include <pathan/XPath2Utils.hpp>
#include <pathan/exceptions/FunctionException.hpp>

#if defined(XERCES_HAS_CPP_NAMESPACE)
XERCES_CPP_NAMESPACE_USE
#endif

using namespace DbXml;
using namespace std;

DbXmlUri::DbXmlUri(const std::string &uri, bool documentUri)
	: valid_(false)
{
	parseUri(0, UTF8ToXMLCh(uri).str(), documentUri);
}

DbXmlUri::DbXmlUri(const XMLCh *uri, bool documentUri)
	: valid_(false)
{
	parseUri(0, uri, documentUri);
}

DbXmlUri::DbXmlUri(const std::string &baseUri, const std::string &relativeUri, bool documentUri)
	: valid_(false)
{
	parseUri(UTF8ToXMLCh(baseUri).str(), UTF8ToXMLCh(relativeUri).str(), documentUri);
}

DbXmlUri::DbXmlUri(const XMLCh *baseUri, const XMLCh *relativeUri, bool documentUri)
	: valid_(false)
{
	parseUri(baseUri, relativeUri, documentUri);
}

// throws if it doesn't exist
XmlContainer DbXmlUri::openContainer(XmlManager &mgr, Transaction *txn) const
{
	Manager &mgrImpl = (Manager &)mgr;

	if (mgrImpl.allowAutoOpen()) {
		if(txn) {
			XmlTransaction childTxn = txn->createChild();
			XmlContainer container(
				mgrImpl.openContainer(
					containerName_, childTxn,
					mgrImpl.getDefaultContainerFlags(),
					mgrImpl.getDefaultContainerType(), 0));
			childTxn.commit();
			return container;
		} else {
			return mgrImpl.openContainer(
				containerName_, 0,
				mgrImpl.getDefaultContainerFlags(),
				mgrImpl.getDefaultContainerType(), 0);
		}
	} else {
		// container must be open
		TransactedContainer *cont =
			mgrImpl.getOpenContainer(containerName_);
		if (cont)
			return cont;
		std::string msg = "Cannot query container: ";
		msg += containerName_;
		msg += ".  Container not open and auto-open is not enabled";
		throw XmlException(
			XmlException::CONTAINER_CLOSED, msg);
	}
}

XmlDocument DbXmlUri::openDocument(XmlManager &mgr, Transaction *txn) const
{
	XmlContainer cont(openContainer(mgr, txn));
	XmlDocument doc(mgr.createDocument());

	OperationContext oc(txn);
	((Container&)cont).getDocument(oc, documentName_, doc, DBXML_LAZY_DOCS);

	return doc;
}

void DbXmlUri::parseUri(const XMLCh *baseUri, const XMLCh *relativeUri, bool documentUri)
{
	try {
		// Parse the URIs
		// (We're happy using the default memory manager for these)
		scoped_ptr<XMLUri> base;
		scoped_ptr<XMLUri> uriObj;

		if(baseUri == 0) {
			uriObj.reset(new XMLUri(relativeUri));
		}
		else {
			base.reset(new XMLUri(baseUri));
			uriObj.reset(new XMLUri(base.get(), relativeUri));
		}

		resolvedUri_ = XMLChToUTF8(uriObj->getUriText()).str();

		/*
		  We are a DbXml URI if our scheme is "dbxml", and we haven't
		  got a host.
		*/
		if(XPath2Utils::equals(Globals::XMLCh_dbxml_scheme, uriObj->getScheme())
		   && (uriObj->getHost() == 0 || (*uriObj->getHost()) == 0)) {
			valid_ = true;

			// The first character is assumed to be a slash
			string path = XMLChToUTF8(uriObj->getPath()).str();
			int pos = -1;
			if(documentUri) pos = path.rfind('/');

			if(pos != -1) {
				containerName_.assign(path, 1, pos - 1);
				documentName_.assign(path, pos + 1, path.size() - pos);
			}
			else {
				containerName_.assign(path, 1, path.size() - 1);
			}
		}
	}
	catch(...)  { // const MalformedURLException&) {
		// don't throw; it's just that the URI isn't valid
		// throw XmlException(XmlException::INTERNAL_ERROR, "Invalid URI format", __FILE__, __LINE__);
	}
}
